@@ -1,7 +1,7 @@
 package TestApp::Controller::Args;
 
 use strict;
-use base 'Catalyst::Base';
+use base 'Catalyst::Controller';
 
 sub args :Local  {
     my ( $self, $c ) = @_;
